package nl.kallestruik.dtweaks.managers

import net.minecraft.core.Direction
import net.minecraft.world.InteractionHand
import net.minecraft.world.entity.Entity
import net.minecraft.world.level.ClipContext
import net.minecraft.world.phys.*
import nl.kallestruik.dlib.ReflectionUtil
import nl.kallestruik.dtweaks.DTweaks
import nl.kallestruik.dtweaks.fakeplayer.FakePlayer
import org.bukkit.Bukkit
import org.bukkit.Location
import org.bukkit.craftbukkit.v1_18_R1.entity.CraftEntity
import org.bukkit.craftbukkit.v1_18_R1.entity.CraftPlayer
import org.bukkit.scheduler.BukkitRunnable
import org.bukkit.scheduler.BukkitTask
import java.util.function.Predicate


/*
    A lot of code in this file is inspired by code from carpet mod.
    Link: https://github.com/gnembon/fabric-carpet
 */
class FakePlayerManager(
    val plugin: DTweaks,
) {

    init {
        // Tick all fake players.
        FakePlayerTickTask().runTaskTimer(plugin, 10, 10)
    }

    fun getPlayerByName(name: String): FakePlayer? {
        val player = Bukkit.getPlayer(name)
        if (player == null || !player.isOnline) return null
        return (player as CraftPlayer).handle as? FakePlayer
    }

    fun toFakePlayer(entity: org.bukkit.entity.Entity): FakePlayer? {
        return (entity as CraftEntity).handle as? FakePlayer
    }

    fun spawnFakePlayer(loc: Location, name: String) {
        val player = Bukkit.getPlayer(name)
        if (player != null && player.isOnline) return
        FakePlayer.atLocation(loc, name)
    }

    fun killFakePlayer(name: String) {
        getPlayerByName(name)?.kill()
    }

    fun killAllFakePlayers() {
        for (player in Bukkit.getOnlinePlayers()) {
            if (!player.isOnline) continue
            val fakePlayer = (player as CraftPlayer).handle as? FakePlayer ?: continue
            fakePlayer.kill()
        }
    }

    private fun tickAll() {
        for (player in Bukkit.getOnlinePlayers()) {
            if (!player.isOnline) continue
            val fakePlayer = (player as CraftPlayer).handle as? FakePlayer ?: continue
            fakePlayer.doTick()
        }
    }

//    fun attackInterval(name: String, interval: Int) {
//
//
//        fakePlayer.swing()
//    }

    fun attack(name: String) {
        val player = getPlayerByName(name) ?: return
        val target = getTarget(player) ?: return

        when (target.type) {
            HitResult.Type.BLOCK -> {/*TODO: Not implemented*/}
            HitResult.Type.ENTITY -> {
                if (target !is EntityHitResult) return

                player.attack(target.entity);
                player.swing(InteractionHand.MAIN_HAND);
                player.resetLastActionTime();
            }
            else -> return
        }
    }

    fun use(name: String) {
        val player = getPlayerByName(name) ?: return
        val target = getTarget(player) ?: return

        for (hand in InteractionHand.values()) {
            when (target.type) {
                HitResult.Type.BLOCK -> {
                    if (target !is BlockHitResult) return

                    player.resetLastActionTime()
                    val world = player.getLevel()
                    val pos = target.blockPos
                    val side = target.direction
                    if (pos.y < player.getLevel().maxBuildHeight - (if (side === Direction.UP) 1 else 0)
                        && world.mayInteract(player, pos)) {
                        val result = player.gameMode.useItemOn(player, world, player.getItemInHand(hand), hand, target)
                        if (result.consumesAction()) {
                            if (result.shouldSwing()) player.swing(hand)
                            return
                        }
                    }
                }
                HitResult.Type.ENTITY -> {
                    if (target !is EntityHitResult) return

                    player.resetLastActionTime()
                    val entity = target.entity
                    val relativeHitPos: Vec3 = target.getLocation().subtract(entity.x, entity.y, entity.z)

                    entity.interactAt(player, relativeHitPos, hand).consumesAction()

                    // fix for SS itemframe always returns CONSUME even if no action is performed
                    player.interactOn(entity, hand).consumesAction()
                }
            }

            player.gameMode.useItem(player, player.getLevel(), player.getItemInHand(hand), hand)
        }
    }

    //
    // Ray tracing code borrowed from carpet mod: https://github.com/gnembon/fabric-carpet/blob/master/src/main/java/carpet/helpers/Tracer.java
    //

    fun getTarget(player: FakePlayer): HitResult? {
        val reach: Double = if (player.gameMode.isCreative) 5.0 else 4.5f.toDouble()
        return rayTrace(player, 1F, reach, false)
    }

    fun rayTrace(source: FakePlayer, partialTicks: Float, reach: Double, fluids: Boolean): HitResult? {
        val blockHit = rayTraceBlocks(source, partialTicks, reach, fluids)
        var maxSqDist = reach * reach
        if (blockHit != null) {
            maxSqDist = blockHit.location.distanceToSqr(source.getEyePosition(partialTicks))
        }
        val entityHit = rayTraceEntities(source, partialTicks, reach, maxSqDist)
        return entityHit ?: blockHit
    }

    fun rayTraceBlocks(source: FakePlayer, partialTicks: Float, reach: Double, fluids: Boolean): BlockHitResult? {
        val pos: Vec3 = source.getEyePosition(partialTicks)
        val rotation: Vec3 = source.getViewVector(partialTicks)
        val reachEnd = pos.add(rotation.x * reach, rotation.y * reach, rotation.z * reach)
        return source.level.clip(
            ClipContext(
                pos,
                reachEnd,
                ClipContext.Block.OUTLINE,
                if (fluids) ClipContext.Fluid.ANY else ClipContext.Fluid.NONE,
                source
            )
        )
    }

    fun rayTraceEntities(source: FakePlayer, partialTicks: Float, reach: Double, maxSqDist: Double): EntityHitResult? {
        val pos = source.getEyePosition(partialTicks)
        val reachVec = source.getViewVector(partialTicks).scale(reach)
        val box = source.boundingBox.expandTowards(reachVec).inflate(1.0)
        return rayTraceEntities(source, pos, pos.add(reachVec), box, maxSqDist) {
                e -> !e.isSpectator && e.isPickable
        }
    }

    fun rayTraceEntities(
        source: FakePlayer,
        start: Vec3,
        end: Vec3,
        box: AABB,
        maxSqDistance: Double,
        predicate: Predicate<Entity>,
        ): EntityHitResult? {
        val world = source.level
        var targetDistance = maxSqDistance
        var target: Entity? = null
        var targetHitPos: Vec3? = null
        for (current in world.getEntities(source, box, predicate)) {
            val currentBox = current.boundingBox.inflate(current.pickRadius.toDouble())
            val currentHit = currentBox.clip(start, end)
            if (currentBox.contains(start)) {
                if (targetDistance >= 0) {
                    target = current
                    targetHitPos = currentHit.orElse(start)
                    targetDistance = 0.0
                }
            } else if (currentHit.isPresent) {
                val currentHitPos: Vec3 = currentHit.get()
                val currentDistance = start.distanceToSqr(currentHitPos)
                if (currentDistance < targetDistance || targetDistance == 0.0) {
                    if (current.rootVehicle === source.rootVehicle) {
                        if (targetDistance == 0.0) {
                            target = current
                            targetHitPos = currentHitPos
                        }
                    } else {
                        target = current
                        targetHitPos = currentHitPos
                        targetDistance = currentDistance
                    }
                }
            }
        }
        return if (target == null) null else EntityHitResult(target, targetHitPos!!)
    }

    inner class FakePlayerTickTask: BukkitRunnable() {
        override fun run() {
            tickAll()
        }

    }
}