package nl.kallestruik.dtweaks.tweaks.miscellaneoustweaks

import com.google.common.base.Preconditions
import net.kyori.adventure.text.Component
import net.kyori.adventure.text.format.NamedTextColor
import net.kyori.adventure.text.format.TextDecoration
import nl.kallestruik.dtweaks.BundleColor
import nl.kallestruik.dtweaks.BundleVariant
import nl.kallestruik.dtweaks.DTweaks
import nl.kallestruik.dtweaks.tweaks.ITweak
import org.bukkit.Bukkit
import org.bukkit.Keyed
import org.bukkit.Material
import org.bukkit.NamespacedKey
import org.bukkit.event.EventHandler
import org.bukkit.event.HandlerList
import org.bukkit.event.Listener
import org.bukkit.event.inventory.InventoryClickEvent
import org.bukkit.event.inventory.PrepareItemCraftEvent
import org.bukkit.inventory.*
import org.bukkit.inventory.meta.BundleMeta
import org.bukkit.persistence.PersistentDataType


class BetterBundles(
    private val plugin: DTweaks
): ITweak, Listener {
    override fun getIdentifier(): String = "BetterBundles"
    override fun getCategories(): List<String> = listOf("miscellaneous", "survival")

    private val recipeKeys = mutableListOf<NamespacedKey>()

    override fun onEnable() {
        plugin.server.pluginManager.registerEvents(this, plugin)

        for (color in BundleColor.values()) {
            val result = ItemStack(Material.BUNDLE)
            val resultMeta = result.itemMeta
            resultMeta.setCustomModelData(BundleVariant.NORMAL.modelDataStart + color.modelData)
            resultMeta.persistentDataContainer.set(COLOR_KEY, PersistentDataType.STRING, color.toString())
            result.itemMeta = resultMeta

            val key = NamespacedKey(plugin, "${color.name.lowercase()}_color_bundle")
            recipeKeys.add(key)

            addDyeRecipe(key, color.dye, result)
        }

        for (variant in BundleVariant.values()) {
            if (variant.material == null)
                continue

            val result = ItemStack(Material.BUNDLE)
            val resultMeta = result.itemMeta
            resultMeta.setCustomModelData(variant.modelDataStart + BundleColor.BROWN.modelData)
            resultMeta.persistentDataContainer.set(VARIANT_KEY, PersistentDataType.STRING, variant.toString())

            result.itemMeta = resultMeta

            val key = NamespacedKey(plugin, "${variant.name.lowercase()}_variant_bundle")
            recipeKeys.add(key)

            addUpgradeRecipe(key, variant.material, result)
        }
    }

    override fun onDisable() {
        HandlerList.unregisterAll(this)
        for (key in recipeKeys) {
            plugin.server.removeRecipe(key)
        }

        recipeKeys.clear()
    }

    val VARIANT_KEY = NamespacedKey(plugin, "variant")
    val COLOR_KEY = NamespacedKey(plugin, "color")

    private fun addDyeRecipe(key: NamespacedKey, dye: Material, result: ItemStack) {
        val recipe = ShapelessRecipe(key, result)
        recipe.addIngredient(dye)
        recipe.addIngredient(Material.BUNDLE)

        plugin.server.addRecipe(recipe)
    }

    private fun addUpgradeRecipe(key: NamespacedKey, material: Material, result: ItemStack) {
        val recipe = ShapedRecipe(key, result)
        recipe.shape("MMM", "MBM", "MMM")
        recipe.setIngredient('M', material)
        recipe.setIngredient('B', Material.BUNDLE)

        plugin.server.addRecipe(recipe)
    }

    private fun isBundleRecipe(recipe: Recipe): Boolean {
        return isDyeRecipe(recipe) || isVariantRecipe(recipe)
    }

    private fun isDyeRecipe(recipe: Recipe): Boolean {
        if (recipe !is Keyed)
            return false

        val key = recipe.key
        if (key.namespace != plugin.getNamespace())
            return false

        if (!key.key.endsWith("_color_bundle"))
            return false

        return true
    }

    private fun isVariantRecipe(recipe: Recipe): Boolean {
        if (recipe !is Keyed)
            return false

        val key = recipe.key
        if (key.namespace != plugin.getNamespace())
            return false

        if (!key.key.endsWith("_variant_bundle"))
            return false

        return true
    }

    @EventHandler
    fun onCraft(event: PrepareItemCraftEvent) {
        val recipe = event.recipe ?: return

        if (!isBundleRecipe(recipe))
            return


        val inventory = event.inventory
        Bukkit.getScheduler().runTask(
            plugin,
            Runnable {
                inventory.result =
                    if (isDyeRecipe(recipe)) getDyeResult(recipe, inventory)
                    else if (isVariantRecipe(recipe)) getVariantResult(recipe, inventory)
                    else null
            })
    }

    private fun getDyeResult(recipe: Recipe, inventory: CraftingInventory): ItemStack? {
        val bundle = inventory.matrix?.filter { it?.type == Material.BUNDLE }?.get(0) ?: return null
        val recipeResult = recipe.result

        val result = bundle.clone()
        val resultMeta = result.itemMeta
        resultMeta.persistentDataContainer.set(
            COLOR_KEY,
            PersistentDataType.STRING,
            recipeResult.itemMeta.persistentDataContainer.get(
                COLOR_KEY,
                PersistentDataType.STRING
            ) ?: BundleColor.BROWN.name
        )

        result.itemMeta = resultMeta
        updateBundle(result)

        return result
    }

    private fun getVariantResult(recipe: Recipe, inventory: CraftingInventory): ItemStack? {
        val bundle = inventory.matrix?.filter { it?.type == Material.BUNDLE }?.get(0) ?: return null
        val recipeResult = recipe.result

        val result = bundle.clone()
        val resultMeta = result.itemMeta
        resultMeta.persistentDataContainer.set(
            VARIANT_KEY,
            PersistentDataType.STRING,
            recipeResult.itemMeta.persistentDataContainer.get(
                VARIANT_KEY,
                PersistentDataType.STRING
            ) ?: BundleVariant.NORMAL.name
        )

        result.itemMeta = resultMeta
        updateBundle(result)

        return result
    }


    @EventHandler
    fun onInventoryClick(event: InventoryClickEvent) {
        if (!event.isRightClick)
            return

        val cursor = event.cursor ?: return
        val clicked = event.currentItem ?: return

        val bundle: ItemStack
        val item: ItemStack
        var itemIsCursor = true

        if (cursor.type == Material.BUNDLE) {
            bundle = cursor
            item = clicked
            itemIsCursor = false
        } else if (clicked.type == Material.BUNDLE) {
            bundle = clicked
            item = cursor
        } else
            return

        if (item.type.isAir || item.amount == 0) {
            val itemStack = removeItemFromBundle(bundle) ?: return

            if (itemIsCursor)
                event.cursor = itemStack
            else
                event.currentItem = itemStack
        } else
            addItemToBundle(bundle, item)

        event.isCancelled = true
    }

    private fun removeItemFromBundle(bundle: ItemStack): ItemStack? {
        val meta = bundle.itemMeta as BundleMeta
        if (!meta.hasItems())
            return null

        val itemsInBundle = mutableListOf<ItemStack>()
        itemsInBundle.addAll(meta.items)
        val toReturn = itemsInBundle.removeFirst()

        meta.setItems(itemsInBundle)
        bundle.itemMeta = meta

        updateBundle(bundle)

        return toReturn
    }

    private fun addItemToBundle(bundle: ItemStack, item: ItemStack) {
        var meta = bundle.itemMeta as BundleMeta
        val persistentContainer = bundle.itemMeta.persistentDataContainer
        val variant = BundleVariant.valueOf(persistentContainer.get(VARIANT_KEY, PersistentDataType.STRING) ?: "NORMAL")

        val currentWeight = computeWeight(meta.items)
        val availableWeight = variant.maxWeight - currentWeight

        val weightPerItem = getWeightPerItem(item)
        val totalStackWeight = weightPerItem * item.amount
        if (totalStackWeight > availableWeight) {
            val toStoreAmount = availableWeight / weightPerItem
            if (toStoreAmount <= 0)
                return

            meta = addItemToBundleProper(meta, item.asQuantity(toStoreAmount))
            item.amount -= toStoreAmount
        } else {
            meta = addItemToBundleProper(meta, item.clone())
            item.amount = 0
        }

        bundle.itemMeta = meta

        updateBundle(bundle)
    }

    fun updateBundle(bundle: ItemStack) {
        val meta = bundle.itemMeta as BundleMeta
        val persistentContainer = meta.persistentDataContainer

        val variant = BundleVariant.valueOf(persistentContainer.get(VARIANT_KEY, PersistentDataType.STRING) ?: "NORMAL")
        val color = BundleColor.valueOf(persistentContainer.get(COLOR_KEY, PersistentDataType.STRING) ?: "BROWN")

        val customModelData = variant.modelDataStart + color.modelData

        meta.setCustomModelData(customModelData)
        meta.addItemFlags(ItemFlag.HIDE_POTION_EFFECTS)
        meta.lore(listOf(Component
            .text("${computeWeight(meta.items)}/${variant.maxWeight}")
            .color(NamedTextColor.GRAY)
            .decoration(TextDecoration.ITALIC, false)))

        bundle.itemMeta = meta
    }

    private fun computeWeight(items: List<ItemStack>): Int {
        return items.sumOf { getWeight(it) }
    }

    private fun getWeight(item: ItemStack): Int {
        return item.amount * getWeightPerItem(item)
    }

    private fun getWeightPerItem(item: ItemStack): Int {
        // Special case for bundle nesting.
        if (item.type == Material.BUNDLE) {
            val meta = item.itemMeta as BundleMeta
            return if (meta.hasItems())
                64
            else
                4
        }

        // Calculate the right weight.
        return (64 / item.maxStackSize)
    }

    private fun addItemToBundleProper(itemMeta: BundleMeta, itemStack: ItemStack): BundleMeta {
        Preconditions.checkArgument(!itemStack.type.isAir, "item is null or air")

        var leftOver = true
        val items = mutableListOf<ItemStack>()
        items.addAll(itemMeta.items)

        for (item in itemMeta.items) {
            if (item.isSimilar(itemStack)) {
                if (item.amount >= item.maxStackSize)
                    continue

                items.remove(item)
                val spaceLeft = item.maxStackSize - item.amount
                if (spaceLeft >= itemStack.amount) {
                    item.amount += itemStack.amount
                    itemStack.amount = 0
                    items.add(0, item)
                    leftOver = false
                    break
                } else {
                    item.amount = item.maxStackSize
                    itemStack.amount -= spaceLeft
                    items.add(0, item)
                }
            }
        }

        if (leftOver)
            items.add(0, itemStack)

        itemMeta.setItems(items)

        return itemMeta
    }
}